package com.apelon.modules.dts.editor.tableviewer;


import com.apelon.beans.dts.plugin.DTSAppManager;
import com.apelon.dts.client.association.AssociationType;
import com.apelon.dts.client.association.ConceptAssociation;
import com.apelon.dts.client.association.Synonym;
import com.apelon.dts.client.attribute.DTSProperty;
import com.apelon.dts.client.attribute.DTSPropertyType;
import com.apelon.dts.client.attribute.DTSRole;
import com.apelon.dts.client.attribute.DTSRoleType;
import com.apelon.dts.client.common.DTSObject;
import com.apelon.dts.client.concept.ConceptAttributeSetDescriptor;
import com.apelon.dts.client.concept.DTSConcept;
import com.apelon.dts.client.concept.OntylogConcept;
import com.apelon.dts.dtsutil.QualifiedObject;
import org.apache.log4j.Category;


/**
 * attr to hold table viewer attribute information
 * Copyright (c) 2006 Apelon, Inc. All rights reserved.
 * @since 3.4.0
 */
class TableViewerAttribute
{
    //definitions for fixed attributes
    final static String CONCEPT_NAME = "Name";
    final static String CONCEPT_CODE = "Code";
    final static String CONCEPT_ID = "Id";
    final static String PREFERRED_NAME = "Pref. Name";
    final static String NAMESPACE = "Namespace";
    final static String INDIRECT_DELIM = "->";

    private final static ConceptAttributeSetDescriptor allCASD =
        ConceptAttributeSetDescriptor.ALL_ATTRIBUTES;

    private final QualifiedObject attr;             //a DTS attribute,
    //object may be null if "hard" concept attribute

    //displayed attr if "indirect" attribute, or null
    private final QualifiedObject dispAttr;

    /** explicit compound attribute */
    TableViewerAttribute(String name, String dispname) {
        this(name+INDIRECT_DELIM+dispname);
    }

    TableViewerAttribute(String name) {
        int idx;
        //have to check for indirect name
        if ((idx = name.indexOf(INDIRECT_DELIM))<0) {
            //direct name
            attr = new QualifiedObject(name);               //may have null object
            dispAttr = null;
        }
        else {
            //may have null object
            attr = new QualifiedObject(name.substring(0, idx));
            //may have null object
            dispAttr = new QualifiedObject(name.substring(idx+2));
        }
    }

    TableViewerAttribute(DTSObject obj) {
        this( new QualifiedObject(obj), null);
    }

    TableViewerAttribute(QualifiedObject attr) {
        this(attr, null);
    }

    TableViewerAttribute(QualifiedObject attr, QualifiedObject disp) {
        this.attr = attr;
        dispAttr = disp;
    }

    QualifiedObject getAttribute() {
        return attr;
    }

    QualifiedObject getDisplayAttribute() {
        return dispAttr;
    }

    boolean isLocal() {
        return attr.isLocal();
    }

    public String toString() {
        return attr.getName()+(dispAttr==null?"":INDIRECT_DELIM+dispAttr.getName());
    }

    public boolean equals(Object obj) {
        if (!(obj instanceof TableViewerAttribute)) return false;
        return toString().equals(obj.toString());
    }

    /** need to overwrite hashCode since equals is overwritten */
    public int hashCode() {
        return toString().hashCode();
    }

    /**
     * tests if attribute is a valid Table Viewer Attribute (has valid object types)
     * @return true if attribute is valid in Namespace
     */
    boolean isValid() {
    	DTSObject obj = attr.getObject();
    	//direct must be fixed, property, role, or association
    	if (obj==null)
    		//if object is name, must be a fixed attribute
        	return isFixedAttribute(attr.getName()) && dispAttr==null;
    	//direct attribute?
    	if (obj instanceof DTSPropertyType) return dispAttr==null;
    	if (!((obj instanceof DTSRoleType) || (obj instanceof AssociationType))) return false;
    	//indirect attribute, validate displaying as fixed or property
    	obj = dispAttr.getObject();
    	if (obj==null) return isFixedAttribute(dispAttr.getName());
    	return obj instanceof DTSPropertyType;
    }
    
    /**
     * test if name is a fixed
     * @param name name of attribute
     * @return
     */
    private boolean isFixedAttribute(String name) {
		return (name.equals(CONCEPT_NAME) || name.equals(CONCEPT_CODE) ||
				name.equals(CONCEPT_ID) || name.equals(PREFERRED_NAME) ||
				name.equals(NAMESPACE));
    }

    /**
     * test if attribute is direct: a fixed attribute or property
     * @return true if attribute is direct
     */
    boolean isDirectAttribute() {
        return isDirectAttribute(attr);
    }

    /**
     * is argument a direct attribute: a fixed attribute or property
     * @param obj a QualifiedObject
     * @return true if attribute is direct
     */
    static boolean isDirectAttribute(QualifiedObject obj) {
        return (obj.getObject()==null ||
                (obj.getObject() instanceof DTSPropertyType));
    }
     
    /****************************************************************/
    //value routines

    /**
     * get a string attribute value; does not support "repeated" values
     * non-existent values are returned as ""
     * @return the string value of the attribute
     */
    String getValue(DTSConcept con) {
        if (isDirectAttribute()) return getDirectValue(attr, con);
        //indirect attribute
        return getDirectValue(dispAttr, getConceptValue(con));
    }

    /**
     * get the scalar value for a concept attr
     * assumes con has all necessary attributes fetched
     * @return the string value of the direct attribute
     */
    private String getDirectValue(QualifiedObject attr, DTSConcept con) {
        DTSProperty[] props;
        int i;

        String value = "";
        String name = attr.getName();
        //System.out.println("getting '"+name+"' for "+con);
        if (con==null) return value;   //a concept value may not have been present
        //if (name.equals(CONCEPT)) return conn.getNameForObject(con);
        if (name.equals(CONCEPT_NAME)) return con.getName();
        if (name.equals(CONCEPT_CODE)) return con.getCode();
        if (name.equals(CONCEPT_ID)) return con.getId()+"";
        if (name.equals(PREFERRED_NAME)) {
            Synonym syn = con.getFetchedPreferredTerm();
            if (syn==null) return con.getName();
            else return syn.getTerm().getName();
        }
        if (name.equals(NAMESPACE)) {
            int id = con.getNamespaceId();
            try {
                return DTSAppManager.getQuery().getNamespaceQuery().
                    findNamespaceById(id).getName();
            } catch (Exception e) {
                Category.getInstance("tableviewer").error(e);
                e.printStackTrace();
                return "*unknown*";
            }
        }

        //System.out.println(nam);
        //got a "real" attribute;
        DTSObject object = attr.getObject();
        if (object instanceof DTSPropertyType) {
            //get matching properties, return first value
            props = con.getFetchedProperties();
            for (i=0; i<props.length; i++)
                if (props[i].getPropertyType().equals(object))
                    //value += (value.equals("")?"":valueDelim)+props[i].getValue();
                    return props[i].getValue();
        }
        return value;
    }

 
    /** get the Concept value of this attr (role or association) 
     * @return the Concept value of the attribute
     */
    private DTSConcept getConceptValue(DTSConcept con) {
        DTSConcept res = null;                //result concept
        DTSRole[] roles;
        ConceptAssociation[] assns;
        int i;

        DTSObject object = attr.getObject();
        if (object instanceof DTSRoleType) {
            roles = ((OntylogConcept)con).getFetchedRoles();
            //get first matching role
            for (i=0; i<roles.length; i++)
                if (roles[i].getRoleType().equals(object)) {
                    res = roles[i].getValueConcept();
                    break;
                }
        }
        else if (object instanceof AssociationType) {
        	if (attr.isInverse()) {
	            assns = con.getFetchedInverseConceptAssociations();
	            //get first matching assn
	            for (i=0; i<assns.length; i++) {
	                if (assns[i].getAssociationType().equals(object)) {
	                    res = assns[i].getFromConcept();
	                    break;
	                }
	            }
        	}
        	else {
	            assns = con.getFetchedConceptAssociations();
	            //get first matching assn
	            for (i=0; i<assns.length; i++) {
	                if (assns[i].getAssociationType().equals(object)) {
	                    res = assns[i].getToConcept();
	                    break;
	                }
	            }
        	}
        }
        if (res!=null) {
            //need to get all concept attributes
            try {
                res = DTSAppManager.getQuery().getSearchQuery().
                    findConceptByName(res.getName(), res.getNamespaceId(), allCASD);
            } catch (Exception e) {
                Category.getInstance("tableviewer").error(e);
                e.printStackTrace();
                res = null;
            }
        }
        return res;
    }
}