package com.apelon.modules.dts.editor.tableviewer;

import java.util.Arrays;
import java.util.Comparator;
import javax.swing.JOptionPane;

import com.apelon.apps.dts.editor.modules.DTSEditorModuleMgr;
import com.apelon.beans.dts.plugin.DTSAppManager;
import com.apelon.dts.client.concept.ConceptAttributeSetDescriptor;
import com.apelon.dts.client.concept.DTSConcept;
import com.apelon.dts.client.events.DataChangeEvent;
import com.apelon.dts.common.subset.Subset;
import com.apelon.dts.client.events.SubsetEvent;
import com.apelon.dts.common.subset.SubsetDescriptor;
import org.apache.log4j.Category;
        

/**
 * selection panel for subsets
 * <p>
 * Copyright (c) 2006 Apelon, Inc. All rights reserved.
 * @since 3.4.0
 */
class TableViewerSubsetPanel extends TableViewerDirectorPanel implements TableViewerTableDirector {
        
    final static ConceptAttributeSetDescriptor allCASD = 
        ConceptAttributeSetDescriptor.ALL_ATTRIBUTES;
    final static int MAX_SUBSET_SIZE = 500;
    
    //the variable below is necessary for subset event handler (can't access deleted subset)
    private int subsetId;						//id of loaded subset
        
    TableViewerSubsetPanel(DTSEditorModuleMgr mgr, TableViewerTablePanel panel) {
        super(mgr, panel, LOAD);
    }
    
    //* get the namespace id for the active subset
    public int getActiveNamespaceId() {
    	try {
    		if (subsetId>=0) {
    			int[] ids = DTSAppManager.getQuery().getSubsetQuery().fetchNamespaces(subsetId);
    			if (ids.length>0) return ids[0];
    		}
	    } catch (Exception e) {
	        Category.getInstance("tableviewer").error(e);
	        e.printStackTrace();
	    }
        return -1;
    }
    
    /**************************************************/
    /** get the label for this panel */
    public String getDirectorLabel() { return "Subset"; }
    
    /** get sorted list of subset names */
    protected String[] getDirectorNames() {
        String[] names;
        Subset[] subsets = getSubsets();
        names = new String[subsets.length];
        for (int i=0; i<subsets.length; i++) names[i] = subsets[i].getName();
        Arrays.sort(names, 0, names.length, new Comparator<String>() { 
                public int compare(String a, String b) {
                    return a.compareToIgnoreCase(b);
                }
        }
                    );
        return names;
    }
         
    /** get an array of the subsets */
    private Subset[] getSubsets() {
        try  {
            return DTSAppManager.getQuery().getSubsetQuery().
                findAll(SubsetDescriptor.DEFAULT_SUBSET_DESCRIPTOR);
        } catch (Exception e) {
            Category.getInstance("tableviewer").error(e);
            e.printStackTrace();
            return new Subset[0];
        }
    }
         
    /** get the subset id for a subset name */
    private int getSubsetId(String name) {
        try  {
            Subset subset = DTSAppManager.getQuery().getSubsetQuery().fetch(name);
            if (subset!=null) return subset.getId();
        } catch (Exception e) {
            Category.getInstance("tableviewer").error(e);
            e.printStackTrace();
        }
        return -2;
    }
    
    /** get a Subset by its id */
    private Subset getSubsetById(int id) {
        try  {
            return DTSAppManager.getQuery().getSubsetQuery().fetch(id);
        } catch (Exception e) {
            Category.getInstance("tableviewer").error(e);
            e.printStackTrace();
        }
        return null;
    }
         
    //methods for Subset Lists
         
    private String getKey(String name) {
        return "{"+name+"}";
    }

    /** load table, 
     * call with element name, load function (LOAD/OPEN)
     * can't use active variables */
     protected boolean loadTable(String name, int func) {
    	boolean done = false;
    	subsetId = -1;
        setCursor(busyCursor);
        try {
            System.out.println("TableViewer.getting concepts for Subset '"+name+"'");
            //get all concepts in selected subset
            int sid = getSubsetId(name);
            if (sid<0) 
                throw new Exception("Invalid Subset name '"+name+"'.");
            Subset subset = queryMgr.getSubsetQuery().fetch(sid);
            int size = queryMgr.getSubsetQuery().fetchConceptCount(subset);
            System.out.println(name+" size="+size+" ("+subset.getConceptCount()+")");
                        
            if (size>MAX_SUBSET_SIZE) {
                //don't allow huge subsets
                JOptionPane.showMessageDialog(null, 
                      "Subset size of "+size+" exceeds Viewer maximum.", 
                      "Table Viewer Error", JOptionPane.ERROR_MESSAGE);
            }
            else {
                DTSConcept[] cons = queryMgr.getSubsetQuery().fetchConcepts(sid);
                //now fill cons with casd
                for (int i=0; i<cons.length; i++)
                    cons[i] = queryMgr.getSearchQuery().
                        findConceptById(cons[i].getId(), 
                                        cons[i].getNamespaceId(), allCASD);
                System.out.println("TableViewer.found "+cons.length+" concepts");
                tablePanel.loadTableViewer(this, getKey(name), cons, 0);    //no editing
                subsetId = sid;
            }
            done = true;
        } catch (Exception e) {
            Category.getInstance("tableviewer").error(e);
            System.out.println("Error loading concepts!");
            e.printStackTrace();
            moduleMgr.showErrorMessage("Error retrieving Subset concepts.");
        }
        setCursor(defCursor);
        return done;
    }
         
    /*******************************************************************************/
    //handle subset event
    void processSubsetEvent(SubsetEvent event) {
        System.out.println("subset action:"+event+"="+event.getAction());
    	if (event.getAction()==SubsetEvent.ACTION_ADD) {
    		System.out.println("subset add:"+getSubsetById(event.getSubsetId()).getName());
    		loadKeyCombo("", "");				//just load the new name
    	}
        if (event.getAction()== SubsetEvent.ACTION_DELETE) {
    		//can't access the subset so have to use saved vars
        	loadKeyCombo(event.getSubsetId()==subsetId?getActiveName():"", "");
        }
        if (event.getAction()==SubsetEvent.ACTION_MODIFY) {
        	//can't rename a subset so this is a content modify
        	String name = getSubsetById(event.getSubsetId()).getName();
    		System.out.println("subset modify:"+name);
    		if (name.equals(getActiveName())) {
    			JOptionPane.showMessageDialog(null, 
    			    "Subset '"+getActiveName()+"' has been modified. Press Load to refresh subset data.", 
                    "Subset Update", JOptionPane.WARNING_MESSAGE);
    		}
        }
    }

    /**************************************************/
    //concept editing not allowed for subset panel so these are null
    
    /** add additional rows to table */
    public void addRows(Object obj) { }
         
    /** add a new Concept */
    public void newConcept() { }
         
    /** remove namespace rows from table */
    public void removeRows(int[] rows) { }
         
    /** delete namespace row concepts */
    public void deleteConcepts(int[] rows) { }
    
    /** clear table */
    public void clear() { }
    
}
